"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.DocumentManager = void 0;
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_client_1 = require("@fluidframework/server-services-client");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const tenant_1 = require("./tenant");
/**
 * Manager to fetch document from Alfred using the internal URL.
 * @internal
 */
class DocumentManager {
    constructor(internalAlfredUrl, tenantManager, documentStaticDataCache) {
        this.internalAlfredUrl = internalAlfredUrl;
        this.tenantManager = tenantManager;
        this.documentStaticDataCache = documentStaticDataCache;
        if (!this.documentStaticDataCache) {
            server_services_telemetry_1.Lumberjack.info("DocumentManager static data cache is undefined, cache will not be used.");
        }
    }
    // eslint-disable-next-line @rushstack/no-new-null
    async readDocument(tenantId, documentId) {
        // Retrieve the document
        const restWrapper = await this.getBasicRestWrapper(tenantId, documentId);
        const document = await restWrapper.get(`/documents/${tenantId}/${documentId}`);
        if (!document) {
            return null;
        }
        if (this.documentStaticDataCache) {
            // Cache the static properties of the document
            const staticProps = DocumentManager.getStaticPropsFromDoc(document);
            const staticPropsKey = DocumentManager.getDocumentStaticKey(documentId);
            await this.documentStaticDataCache.set(staticPropsKey, JSON.stringify(staticProps));
        }
        // Return the original document that was retrieved
        return document;
    }
    async readStaticProperties(tenantId, documentId) {
        // If the cache is undefined, fetch the document from the database
        if (!this.documentStaticDataCache) {
            server_services_telemetry_1.Lumberjack.verbose("Falling back to database after attempting to read cached static document data, because the DocumentManager cache is undefined.", (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
            const document = (await this.readDocument(tenantId, documentId)) ?? undefined;
            return document;
        }
        // Retrieve cached static document props
        const staticPropsKey = DocumentManager.getDocumentStaticKey(documentId);
        const staticPropsStr = (await this.documentStaticDataCache.get(staticPropsKey)) ?? undefined;
        // If there are no cached static document props, fetch the document from the database
        if (!staticPropsStr) {
            server_services_telemetry_1.Lumberjack.verbose("Falling back to database after attempting to read cached static document data.", (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
            const document = await this.readDocument(tenantId, documentId);
            if (!document) {
                server_services_telemetry_1.Lumberjack.warning("Fallback to database failed, document not found.", (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
                return undefined;
            }
            return DocumentManager.getStaticPropsFromDoc(document);
        }
        // Return the static data, parsed into a JSON object
        const staticProps = JSON.parse(staticPropsStr);
        return staticProps;
    }
    async purgeStaticCache(documentId) {
        // If the cache is undefined, do nothing, because there are no cached static properties to purge
        if (!this.documentStaticDataCache) {
            server_services_telemetry_1.Lumberjack.error("Cannot purge document static properties cache, because the DocumentManager cache is undefined.");
            return;
        }
        if (this.documentStaticDataCache.delete === undefined) {
            server_services_telemetry_1.Lumberjack.error("Cannot purge document static properties cache, because the cache does not have a delete function.");
            return;
        }
        const staticPropsKey = DocumentManager.getDocumentStaticKey(documentId);
        await this.documentStaticDataCache.delete(staticPropsKey);
    }
    async getBasicRestWrapper(tenantId, documentId) {
        const scopes = [protocol_definitions_1.ScopeType.DocRead];
        const accessToken = await this.tenantManager.signToken(tenantId, documentId, scopes);
        const getDefaultHeaders = () => {
            return {
                Authorization: `Basic ${accessToken}`,
            };
        };
        const refreshTokenIfNeeded = (0, tenant_1.getRefreshTokenIfNeededCallback)(this.tenantManager, documentId, tenantId, scopes, "documentManager");
        const restWrapper = new server_services_client_1.BasicRestWrapper(this.internalAlfredUrl, undefined /* defaultQueryString */, undefined /* maxBodyLength */, undefined /* maxContentLength */, getDefaultHeaders(), undefined /* Axios */, undefined /* refreshDefaultQueryString */, getDefaultHeaders /* refreshDefaultHeaders */, () => (0, server_services_telemetry_1.getGlobalTelemetryContext)().getProperties().correlationId /* getCorrelationId */, () => (0, server_services_telemetry_1.getGlobalTelemetryContext)().getProperties() /* getTelemetryContextProperties */, refreshTokenIfNeeded /* refreshTokenIfNeeded */);
        return restWrapper;
    }
    /**
     * Creates a cache key to retreive static data from a document
     *
     * @param documentId - ID of the document to create an access key for
     * @returns - A cache key to access static data for [documentId]
     */
    static getDocumentStaticKey(documentId) {
        return `staticData:${documentId}`;
    }
    /**
     * Extracts the static properties from an IDocument
     *
     * @param document - Document to get properties from
     * @returns - The static properties of [document]
     */
    static getStaticPropsFromDoc(document) {
        return {
            version: document.version,
            createTime: document.createTime,
            documentId: document.documentId,
            tenantId: document.tenantId,
            storageName: document.storageName,
            isEphemeralContainer: document.isEphemeralContainer,
        };
    }
}
exports.DocumentManager = DocumentManager;
//# sourceMappingURL=documentManager.js.map