"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.RedisCache = void 0;
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
/**
 * Redis based cache redisClientConnectionManager.getRedisClient()
 * @internal
 */
class RedisCache {
    constructor(redisClientConnectionManager, parameters) {
        this.redisClientConnectionManager = redisClientConnectionManager;
        this.expireAfterSeconds = 60 * 60 * 24;
        this.prefix = "page";
        if (parameters?.expireAfterSeconds) {
            this.expireAfterSeconds = parameters.expireAfterSeconds;
        }
        if (parameters?.prefix) {
            this.prefix = parameters.prefix;
        }
        redisClientConnectionManager.addErrorHandler(undefined, // lumber properties
        "Error with Redis");
    }
    async delete(key) {
        try {
            await this.redisClientConnectionManager.getRedisClient().del(this.getKey(key));
            return true;
        }
        catch (error) {
            const newError = { name: error?.name, message: error?.message };
            server_services_telemetry_1.Lumberjack.error(`Error deleting from cache.`, undefined, newError);
            return false;
        }
    }
    // eslint-disable-next-line @rushstack/no-new-null
    async get(key) {
        try {
            // eslint-disable-next-line @typescript-eslint/return-await
            return this.redisClientConnectionManager.getRedisClient().get(this.getKey(key));
        }
        catch (error) {
            const newError = { name: error?.name, message: error?.message };
            server_services_telemetry_1.Lumberjack.error(`Error getting ${key.substring(0, 20)} from cache.`, undefined, newError);
            throw newError;
        }
    }
    async set(key, value, expireAfterSeconds) {
        try {
            const result = await this.redisClientConnectionManager
                .getRedisClient()
                .set(this.getKey(key), value, "EX", expireAfterSeconds ?? this.expireAfterSeconds);
            if (result !== "OK") {
                throw new Error(result);
            }
        }
        catch (error) {
            const newError = { name: error?.name, message: error?.message };
            server_services_telemetry_1.Lumberjack.error(`Error setting ${key.substring(0, 20)} in cache.`, undefined, newError);
            throw newError;
        }
    }
    async incr(key) {
        try {
            // eslint-disable-next-line @typescript-eslint/return-await
            return this.redisClientConnectionManager.getRedisClient().incr(key);
        }
        catch (error) {
            const newError = { name: error?.name, message: error?.message };
            server_services_telemetry_1.Lumberjack.error(`Error while incrementing counter for ${key.substring(0, 20)} in redis.`, undefined, newError);
            throw newError;
        }
    }
    async decr(key) {
        try {
            // eslint-disable-next-line @typescript-eslint/return-await
            return this.redisClientConnectionManager.getRedisClient().decr(key);
        }
        catch (error) {
            const newError = { name: error?.name, message: error?.message };
            server_services_telemetry_1.Lumberjack.error(`Error while decrementing counter for ${key.substring(0, 20)} in redis.`, undefined, newError);
            throw newError;
        }
    }
    /**
     * Translates the input key to the one we will actually store in redis
     */
    getKey(key) {
        return `${this.prefix}:${key}`;
    }
}
exports.RedisCache = RedisCache;
//# sourceMappingURL=redis.js.map