/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { ISignalClient } from "@fluidframework/protocol-definitions";
import { ICollaborationSessionClient, ICollaborationSession, ICollaborationSessionManager, ICollaborationSessionTracker, IClientManager } from "@fluidframework/server-services-core";
/**
 * Tracks the state of collaboration sessions and manages their lifecycle.
 * @internal
 */
export declare class CollaborationSessionTracker implements ICollaborationSessionTracker {
    /**
     * Client manager used to manage the set of connected clients.
     *
     * @remarks
     * In a multi-service-instance environment, this should track state across all instances.
     */
    private readonly clientManager;
    /**
     * Session manager used to manage the set of active sessions.
     *
     * @remarks
     * In a multi-service-instance environment, this should track state across all instances.
     * This is also used to periodically prune sessions whose timers have expired but were not
     * cleaned up due to a service shutdown or other error.
     */
    private readonly sessionManager;
    /**
     * Timeout in milliseconds after which a session is considered inactive and can be ended.
     * After a the last client in a document's session disconnects, this countdown begins.
     * Default: 10 minutes
     */
    private readonly sessionActivityTimeoutMs;
    /**
     * Map of session timers keyed by session ID.
     *
     * @remarks
     * When a session is "ended" (last connected client session ends), a timer is started to
     * end the document's collaboration session after a period of inactivity.
     * This map tracks those timers so they can be cleared if a client reconnects.
     * Additionally, when the timer expires, the cross-instance client manager is checked to verify
     * that no clients have reconnected to the session, in which case the timer on this instance is ignored.
     */
    private readonly sessionEndTimers;
    constructor(
    /**
     * Client manager used to manage the set of connected clients.
     *
     * @remarks
     * In a multi-service-instance environment, this should track state across all instances.
     */
    clientManager: IClientManager, 
    /**
     * Session manager used to manage the set of active sessions.
     *
     * @remarks
     * In a multi-service-instance environment, this should track state across all instances.
     * This is also used to periodically prune sessions whose timers have expired but were not
     * cleaned up due to a service shutdown or other error.
     */
    sessionManager: ICollaborationSessionManager, 
    /**
     * Timeout in milliseconds after which a session is considered inactive and can be ended.
     * After a the last client in a document's session disconnects, this countdown begins.
     * Default: 10 minutes
     */
    sessionActivityTimeoutMs?: number);
    startClientSession(client: ICollaborationSessionClient, sessionId: Pick<ICollaborationSession, "tenantId" | "documentId">, knownConnectedClients?: ISignalClient[]): void;
    private startClientSessionCore;
    endClientSession(client: ICollaborationSessionClient, sessionId: Pick<ICollaborationSession, "tenantId" | "documentId">, knownConnectedClients?: ISignalClient[]): void;
    private endClientSessionCore;
    pruneInactiveSessions(): void;
    private pruneInactiveSessionsCore;
    private getSessionAndClients;
    private handleClientSessionTimeout;
    private cleanupSessionOnEnd;
    private getSessionTimerKey;
}
//# sourceMappingURL=sessionTracker.d.ts.map